'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _base = require('../serializers/base-64');

var _base2 = _interopRequireDefault(_base);

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _types = require('../utils/types');

var _types2 = _interopRequireDefault(_types);

var _text = require('./text');

var _text2 = _interopRequireDefault(_text);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/**
 * Debugger.
 *
 * @type {Function}
 */

var debug = (0, _debug2.default)('slate:node');

/**
 * Node.
 *
 * @type {Component}
 */

var Node = function (_React$Component) {
  _inherits(Node, _React$Component);

  function Node() {
    var _Object$getPrototypeO;

    var _temp, _this, _ret;

    _classCallCheck(this, Node);

    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return _ret = (_temp = (_this = _possibleConstructorReturn(this, (_Object$getPrototypeO = Object.getPrototypeOf(Node)).call.apply(_Object$getPrototypeO, [this].concat(args))), _this), _initialiseProps.call(_this), _temp), _possibleConstructorReturn(_this, _ret);
  }

  /**
   * Debug.
   *
   * @param {String} message
   * @param {Mixed} ...args
   */

  /**
   * Should the node update?
   *
   * @param {Object} props
   * @param {Object} state
   * @return {Boolean}
   */

  /**
   * On drag start, add a serialized representation of the node to the data.
   *
   * @param {Event} e
   */

  /**
   * Render.
   *
   * @return {Element} element
   */

  /**
   * Render a `child` node.
   *
   * @param {Node} child
   * @return {Element} element
   */

  /**
   * Render an element `node`.
   *
   * @return {Element} element
   */

  /**
   * Render a text node.
   *
   * @return {Element} element
   */

  return Node;
}(_react2.default.Component);

/**
 * Export.
 */

Node.propTypes = {
  editor: _react2.default.PropTypes.object.isRequired,
  node: _react2.default.PropTypes.object.isRequired,
  renderDecorations: _react2.default.PropTypes.func.isRequired,
  renderMark: _react2.default.PropTypes.func.isRequired,
  renderNode: _react2.default.PropTypes.func.isRequired,
  state: _react2.default.PropTypes.object.isRequired
};
Node.defaultProps = {
  style: {}
};

var _initialiseProps = function _initialiseProps() {
  var _this2 = this;

  this.debug = function (message) {
    for (var _len2 = arguments.length, args = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
      args[_key2 - 1] = arguments[_key2];
    }

    var node = _this2.props.node;
    var key = node.key;
    var kind = node.kind;
    var type = node.type;

    var id = kind == 'text' ? key + ' (' + kind + ')' : key + ' (' + type + ')';
    debug.apply(undefined, [message, '' + id].concat(args));
  };

  this.shouldComponentUpdate = function (props) {
    return props.node != _this2.props.node || props.state.isFocused && props.state.selection.hasEdgeIn(props.node);
  };

  this.onDragStart = function (e) {
    var node = _this2.props.node;

    var encoded = _base2.default.serializeNode(node);
    var data = e.nativeEvent.dataTransfer;
    data.setData(_types2.default.NODE, encoded);

    _this2.debug('onDragStart', e);
  };

  this.render = function () {
    _this2.debug('render');

    var node = _this2.props.node;

    return node.kind == 'text' ? _this2.renderText() : _this2.renderElement();
  };

  this.renderNode = function (child) {
    var _props = _this2.props;
    var editor = _props.editor;
    var renderDecorations = _props.renderDecorations;
    var renderMark = _props.renderMark;
    var renderNode = _props.renderNode;
    var state = _props.state;

    return _react2.default.createElement(Node, {
      key: child.key,
      node: child,
      state: state,
      editor: editor,
      renderDecorations: renderDecorations,
      renderMark: renderMark,
      renderNode: renderNode
    });
  };

  this.renderElement = function () {
    var _props2 = _this2.props;
    var editor = _props2.editor;
    var node = _props2.node;
    var renderNode = _props2.renderNode;
    var state = _props2.state;

    var Component = renderNode(node);
    var children = node.nodes.map(function (child) {
      return _this2.renderNode(child);
    }).toArray();

    // Attributes that the developer must to mix into the element in their
    // custom node renderer component.
    var attributes = {
      'data-key': node.key,
      'onDragStart': _this2.onDragStart
    };

    // If it's a block node with inline children, add the proper `dir` attribute
    // for text direction.
    if (node.kind == 'block' && node.nodes.first().kind != 'block') {
      var direction = node.getTextDirection();
      if (direction == 'rtl') attributes.dir = 'rtl';
    }

    return _react2.default.createElement(
      Component,
      {
        attributes: attributes,
        key: node.key,
        editor: editor,
        node: node,
        state: state
      },
      children
    );
  };

  this.renderText = function () {
    var _props3 = _this2.props;
    var node = _props3.node;
    var editor = _props3.editor;
    var renderDecorations = _props3.renderDecorations;
    var renderMark = _props3.renderMark;
    var state = _props3.state;

    return _react2.default.createElement(_text2.default, {
      key: node.key,
      editor: editor,
      node: node,
      renderDecorations: renderDecorations,
      renderMark: renderMark,
      state: state
    });
  };
};

exports.default = Node;